(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* This is an implemention of a ternary search tree, as
   described in:
   Dr. Dobb's Journal, April 1998, "Ternary Search Trees"
   It's available online at:
   http://www.ddj.com/articles/1998/9804/9804a/9804a.htm
*)
structure TernaryTree (* :> TERNARY_TREE *) =
  struct
    type index = string
    type dimensions = index list
    datatype traversal = Arbitrary | Alphabetical

    (* The datatype for ternary trees containing data *)
    datatype 'a t =
      Leaf of 'a option
    | Node of Char.char * ('a t) * ('a t) * ('a t) * ('a option)

    exception InternalError
    exception NotImplemented

    (* Supposed to be the default and thus most efficient
       traversal order for the collection datastructure. *)
    val defaultTraversal = Arbitrary

    (* An empty tree *)
    fun empty () =
          Leaf NONE

    (* fLookup : 'a t -> (char * substring) option -> 'a option *)
    
    (* The first 2 pattern cases covers when there are no more characters
       in the key string. Then we have found our data - if there is
       any (remember data is of type 'a option) *)
    fun fLookup (Leaf data) NONE = data
      | fLookup (Node (_, _, _, _, data)) NONE = data

      (* End of tree, but more key characters => nothing found *)
      | fLookup (Leaf _) (SOME _) = NONE

      (* More key characters and more tree data => continue search *)
      | fLookup (Node (foundC, lo, e, hi, data)) (s as (SOME(c, cs))) =
          if Char.< (c, foundC) then
            fLookup lo s
          else if Char.> (c, foundC) then
            fLookup hi s
          else
            fLookup e (Substring.getc cs)

    fun lookup table key =
          fLookup table (Substring.getc (Substring.all key))

    (* Lookup and get all values for keys with the given prefix *)
    fun fLookupKeyedPrefix (Leaf NONE) NONE _ acc =
          acc
      | fLookupKeyedPrefix (Leaf (SOME data)) NONE accKey acc =
          (accKey, data)::acc
      | fLookupKeyedPrefix (Node (foundC, lo, e, hi, data)) NONE accKey acc =
          (* We found our prefix - return everything we found *)
          let
            val acc = case data of
                        NONE =>
                          acc
                      | SOME d =>
                          (accKey, d) :: acc
          in
            fLookupKeyedPrefix lo NONE accKey
              (* FIXME: ((str foundC) ^ accKey) gives O(n^2) time complexity! *)
              (fLookupKeyedPrefix e NONE ((str foundC) ^ accKey)
                 (fLookupKeyedPrefix hi NONE accKey acc))
          end
      | fLookupKeyedPrefix (Leaf _) (SOME _) _ acc =
          (* End of tree, but more key characters => nothing found *)
          acc
      | fLookupKeyedPrefix (Node (foundC, lo, e, hi, data)) (s as (SOME(c, cs))) accKey acc =
          (* More key characters and more tree data => continue search *)
          if Char.< (c, foundC) then
            fLookupKeyedPrefix lo s accKey acc
          else if Char.> (c, foundC) then
            fLookupKeyedPrefix hi s accKey acc
          else
            (* FIXME: ((str c) ^ accKey) gives O(n^2) time complexity! *)
            fLookupKeyedPrefix e (Substring.getc cs) ((str c) ^ accKey) acc

    fun lookupKeyedPrefix table prefix =
          fLookupKeyedPrefix table (Substring.getc (Substring.all prefix)) "" nil

    fun lookupPrefix table prefix =
          List.map #2 (lookupKeyedPrefix table prefix)

    fun keysWithPrefix table prefix =
          List.map #1 (lookupKeyedPrefix table prefix)

    (* fInsert : 'a t -> (char * substring) option -> 'a -> 'a t *)

    (* As with fLookup, the first 2 pattern cases convers when there
       are no more characters in the key string. This is where we
       insert the data. Notice that we just overwrite any previous data. *)
    fun fInsert (Leaf _) NONE newData =
          Leaf (SOME(newData))
      | fInsert (Node (foundC, lo, e, hi, _)) NONE newData =
          Node (foundC, lo, e, hi, SOME(newData))

      (* End of tree, but more key characters => insert data in middle subtree *)
      | fInsert (Leaf data) (SOME(c, cs)) newData =
          Node (c, empty (),
                   fInsert (empty ()) (Substring.getc cs) newData,
                   empty (), data)

      (* More key characters and more tree data => continue insert *)
      | fInsert (Node (foundC, lo, e, hi, data)) (s as SOME(c, cs)) newData =
          if Char.< (c, foundC) then
            Node (foundC, fInsert lo s newData, e, hi, data)
          else if Char.> (c, foundC) then
            Node (foundC, lo, e, fInsert hi s newData, data)
          else
            Node (foundC, lo,
                          fInsert e (Substring.getc cs) newData,
                          hi, data)

    fun insert table key data =
          fInsert table (Substring.getc (Substring.all key)) data

    fun fDelete (Leaf _) NONE =
          Leaf NONE
      | fDelete (Node (foundC, lo, e, hi, _)) NONE =
          Node (foundC, lo, e, hi, NONE)

      (* End of tree, but more key characters => no data found *)
      | fDelete (Leaf data) (SOME(c, cs)) =
          Leaf data

      (* More key characters and more tree data => continue delete *)
      | fDelete (Node (foundC, lo, e, hi, data)) (s as SOME(c, cs)) =
          if Char.< (c, foundC) then
            Node (foundC, fDelete lo s, e, hi, data)
          else if Char.> (c, foundC) then
            Node (foundC, lo, e, fDelete hi s, data)
          else
            Node (foundC, lo,
                          fDelete e (Substring.getc cs),
                          hi, data)

    fun delete table key =
          fDelete table (Substring.getc (Substring.all key))

    fun substitute table key f =
          (* This can obviously be done more efficiently, but we
             keep it simple and slightly less efficient for now... *)
          case lookup table key of
            NONE =>
              raise Subscript
          | SOME v =>
              insert table key (f v)

    fun getKeyIfData key NONE =
          nil
      | getKeyIfData key _ =
          [key]

    fun getKeysF prefix (Node (c, lo, e, hi, data)) =
          (getKeysF prefix lo) @
          (getKeyIfData prefix data) @
          (getKeysF (prefix ^ (str c)) e) @
          (getKeysF prefix hi)
      | getKeysF prefix (Leaf data) =
          getKeyIfData prefix data

    fun dimensions table =
          getKeysF "" table

    fun tabulate (indices, tabF) =
          foldl (fn (index, acc) => insert acc index (tabF index))
                (empty ()) indices

    fun tabulateTraversal Arbitrary args =
          tabulate args
      | tabulateTraversal Alphabetical (indices, tabF) =
          (* This would involve sorting the indices
             and then running the normal tabulate. *)
          raise NotImplemented

    fun create (indices, value) =
          tabulate (indices, fn _ => value)
    
    fun sub (aTable, index) =
          case lookup aTable index of
            NONE =>
              raise Subscript
          | SOME v =>
              v
    
    fun mapiRegionF f base (index, acc) =
          case lookup base index of
            NONE =>
              raise InternalError
          | SOME v =>
              insert acc index (f (index, v))

    fun mapiRegion f {base, dimensions} =
          List.foldl (mapiRegionF f base) (empty ()) dimensions

    fun mapi f aTable =
          (* Can obviously be done more efficiently *)
          mapiRegion f {base = aTable, dimensions = dimensions aTable}
      
    fun map f aTable =
          mapi (fn (_, v) => f v) aTable
    
    fun mapTraversal Arbitrary f aTable =
          map f aTable
      | mapTraversal Alphabetical f aTable =
          raise NotImplemented

    fun mapiTraversal Arbitrary f aTable =
          mapi f aTable
      | mapiTraversal Alphabetical f aTable =
          raise NotImplemented

    fun appendF fromTable (addIndex, accTable) =
          case lookup fromTable addIndex of
            NONE =>
              raise InternalError
          | SOME v =>
              insert accTable addIndex v

    fun append (aTable, addTable) =
          (* Could most likely be done more efficiently *)
          let
            val addIndices = dimensions addTable
          in
            foldl (appendF addTable) aTable addIndices
          end

    fun rev aTable =
          (* This operation does not make much sense for a
             ternary tree, so we don't do anything *)
          aTable

    fun foldliF f aTable (index, acc) =
          case lookup aTable index of
            NONE =>
              raise InternalError
          | SOME v =>
              f (index, v, acc)
              
    fun foldli f init aTable =
          (* Can obviously be done more efficiently *)
          List.foldl (foldliF f aTable) init (dimensions aTable)

    fun foldl f init aTable =
          (* Can obviously be done more efficiently *)
          foldli (fn (_, v, acc) => f (v, acc)) init aTable

    fun foldlTraversal Arbitrary f init aTable =
          foldl f init aTable
      | foldlTraversal Alphabetical f init aTable =
          raise NotImplemented

    fun foldliTraversal Arbitrary f init aTable =
          foldli f init aTable
      | foldliTraversal Alphabetical f init aTable =
          raise NotImplemented

    fun appiF f aTable index =
          case lookup aTable index of
            NONE =>
              raise InternalError
          | SOME v =>
              f (index, v)
              
    fun appi f aTable =
          (* Can obviously be done more efficiently *)
          List.app (appiF f aTable) (dimensions aTable)

    fun app f aTable =
          (* Can obviously be done more efficiently *)
          appi (fn (_, v) => f v) aTable

    fun appTraversal Arbitrary f aTable =
          app f aTable
      | appTraversal Alphabetical f aTable =
          raise NotImplemented

    fun appiTraversal Arbitrary f aTable =
          appi f aTable
      | appiTraversal Alphabetical f aTable =
          raise NotImplemented

    fun findF f (e, NONE) =
          if f e then
            SOME e
          else
            NONE
      | findF f (_, acc as (SOME _)) =
          acc

    fun find f aTable =
          foldl (findF f) NONE aTable
    
    fun exists f aTable =
          foldl (fn (e, acc) => acc orelse (f e)) false aTable

    fun toList aTable =
          foldl op:: nil aTable

    fun fromList l =
          (* Maybe this doesn't make much sense. We would have
             to invent the keys, but we _could_ do that... *)
          raise NotImplemented
  end